from __future__ import annotations
from dataclasses import dataclass
import ipaddress
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpHex,
    XmpInt,
    XmpIPv4Address,
    XmpIPv6Address,
    XmpSequence,
    Hex,
)
from .enums import (
    OnOff,
    YesNo,
    ProtocolOption,
    FilterUse,
    InfoAction,
    L2PlusPresent,
    L3Present,
    FilterMode,
    FilterType,
    FilterVlanType,
)


@register_command
@dataclass
class PEF_INIT:
    """
    Prepares for setting up a filter definition.  When called, all filter
    definitions in the shadow-set which are not applied are discarded and replaced
    with the default values (DEFAULT).

    .. note::

        There are 2 register copies used to configure the filters:

            (1) ``Shadow-copy (type value = 0)`` temporary copy configured by sever.
                Values stored in ``shadow-copy`` have no immediate effect on the flow filters. PEF_APPLY will pass the values from the ``shadow-copy`` to the ``working-copy``.

            (2) ``Working-copy (type value = 1)`` reflects what is currently used for filtering in the FPGA.
                ``Working-copy`` cannot be written directly. Only ``shadow-copy`` allows direct write.

            (3) All ``set`` actions are performed on ``shadow-copy`` ONLY.

            (4) Only when PEF_APPLY is called, ``working-copy`` and FPGA are updated with values from the ``shadow-copy``.
    """

    code: typing.ClassVar[int] = 1700
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Setting up a filter definition.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PEF_APPLY:
    """
    Applies filter definitions from "shadow-copy" to "working-copy". This
    also pushes these settings to the FPGA.
    """

    code: typing.ClassVar[int] = 1701
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Applies filter definitions from "shadow-copy" to "working-copy"."""

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PEF_ENABLE:
    """
    Defines if filtering is enabled for the flow.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1702
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        state: OnOff = field(XmpByte())
        """coded byte, specifies the state of the filter."""

    class SetDataAttr(RequestBodyStruct):
        state: OnOff = field(XmpByte())
        """coded byte, specifies the state of the filter."""

    def get(self) -> Token[GetDataAttr]:
        """Get if filtering is enabled for the flow.

        :return: filter state
        :rtype: PEF_ENABLE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, state: OnOff) -> Token[None]:
        """Set the filter state.

        :param state: state of the filter
        :type state: OnOff
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], state=state))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Set the filter state to OFF.
    """

    set_on = functools.partialmethod(set, OnOff.ON)
    """Set the filter state to ON.
    """


@register_command
@dataclass
class PEF_ETHSETTINGS:
    """
    Defines what filter action is performed on the Ethernet header.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1703
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of Ethernet information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of Ethernet information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of Ethernet information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of Ethernet information."""

    def get(self) -> Token[GetDataAttr]:
        """Get the filter action settings on Ethernet header.

        :return: Filter setting
        :rtype: PEF_ETHSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set the filter action settings on Ethernet header.

        :param use: specifies if Ethernet information is expected
        :type use: FilterUse
        :param action: specifies the use of Ethernet information.
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_ETHSRCADDR:
    """
    Defines the Ethernet Source Address settings for the Ethernet filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1704
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of Ethernet Source Address information."""
        value: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the six bytes of the address. Default value: 0x000000000000."""
        mask: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF."""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of Ethernet Source Address information."""
        value: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the six bytes of the address. Default value: 0x000000000000."""
        mask: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF."""

    def get(self) -> Token[GetDataAttr]:
        """Get the Ethernet Source Address settings for the Ethernet filter.

        :return: the Ethernet Source Address settings for the Ethernet filter
        :rtype: PEF_ETHSRCADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: Hex, mask: Hex) -> Token[None]:
        """Set the Ethernet Source Address settings for the Ethernet filter.

        :param use: specifies the use of Ethernet Source Address information
        :type use: OnOff
        :param value: specifying the six bytes of the address. Default value: 0x000000000000.
        :type value: Hex
        :param mask: specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF.
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Ethernet Source Address is not used for the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """Ethernet Source Address is used for the filter."""


@register_command
@dataclass
class PEF_ETHDESTADDR:
    """
    Defines the Ethernet Destination Address settings for the Ethernet filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1705
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of Ethernet information"""
        value: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the six bytes of the address. Default value: 0x000000000000"""
        mask: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of Ethernet information"""
        value: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the six bytes of the address. Default value: 0x000000000000"""
        mask: Hex = field(XmpHex(size=6))
        """six hex bytes, specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the Ethernet Destination Address settings for the Ethernet filter.

        :return: the Ethernet Destination Address settings for the Ethernet filter.
        :rtype: PEF_ETHDESTADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: Hex, mask: Hex) -> Token[None]:
        """Set the Ethernet Destination Address settings for the Ethernet filter.

        :param use: specifies the use of Ethernet Destination Address information
        :type use: OnOff
        :param value: specifying the six bytes of the address. Default value: 0x000000000000
        :type value: Hex
        :param mask: specifying the mask corresponding to the address. Default value: 0xFFFFFFFFFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Ethernet Destination Address is not used for the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """Ethernet Destination Address is used for the filter."""


@register_command
@dataclass
class PEF_L2PUSE:
    """
    Defines what Layer 2+ protocols that are present and may be used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1706
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: L2PlusPresent = field(XmpByte())
        """coded byte, specifies the presence of Layer 2+ protocols."""

    class SetDataAttr(RequestBodyStruct):
        use: L2PlusPresent = field(XmpByte())
        """coded byte, specifies the presence of Layer 2+ protocols."""

    def get(self) -> Token[GetDataAttr]:
        """Get the Layer 2+ protocols settings for the filter.

        :return: the Layer 2+ protocols settings for the filter
        :rtype: PEF_L2PUSE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: L2PlusPresent) -> Token[None]:
        """Set the Layer 2+ protocols settings for the filter.

        :param use:  specifies the presence of Layer 2+ protocols.
        :type use: L2PlusPresent
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use))

    set_na = functools.partialmethod(set, L2PlusPresent.NA)
    """Set the presence of Layer 2+ protocols to NA."""

    set_vlan1 = functools.partialmethod(set, L2PlusPresent.VLAN1)
    """Set the presence of Layer 2+ protocols to one VLAN Tag."""

    set_vlan2 = functools.partialmethod(set, L2PlusPresent.VLAN2)
    """Set the presence of Layer 2+ protocols to two VLAN Tags."""

    set_mpls = functools.partialmethod(set, L2PlusPresent.MPLS)
    """Set the presence of Layer 2+ protocols to MPLS."""


@register_command
@dataclass
class PEF_VLANSETTINGS:
    """
    Defines what filter action is performed on the VLAN header.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1707
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies if VLAN information is expected."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the use of VLAN information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies if VLAN information is expected."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the use of VLAN information."""

    def get(self) -> Token[GetDataAttr]:
        """Get filter action settings on VLAN header.

        :return: filter action settings on VLAN header
        :rtype: PEF_VLANSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set filter action settings on VLAN header.

        :param use: specifies if VLAN information is expected
        :type use: FilterUse
        :param action: specifies the action of VLAN information
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_VLANTAG:
    """
    Basic mode only. Defines the VLAN TAG settings for the VLAN filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1708
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType
    _vlan_type: FilterVlanType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of VLAN information"""
        value: int = field(XmpInt())
        """decimal digits, specifying the 12 bit value of the tag. Default value: 0."""
        mask: Hex = field(XmpHex(size=2))
        """two hex digits, specifying the 12 bit value of the tag. Default value: 0x0FFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of VLAN information"""
        value: int = field(XmpInt())
        """decimal digits, specifying the 12 bit value of the tag. Default value: 0."""
        mask: Hex = field(XmpHex(size=2))
        """two hex digits, specifying the 12 bit value of the tag. Default value: 0x0FFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the VLAN TAG settings for the VLAN filter.

        :return: the VLAN TAG settings for the VLAN filter
        :rtype: PEF_VLANTAG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._vlan_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set the VLAN TAG settings for the VLAN filter.

        :param use: specifies the use of VLAN TAG information
        :type use: OnOff
        :param value: specifying the 12 bit value of the tag. Default value: 0.
        :type value: int
        :param mask: specifying the 12 bit value of the tag. Default value: 0x0FFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._vlan_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """VLAN TAG is not used for the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """VLAN TAG is used for the filter."""


@register_command
@dataclass
class PEF_VLANPCP:
    """
    Basic mode only. Defines the VLAN PCP settings for the VLAN filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1709
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType
    _vlan_type: FilterVlanType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of VLAN information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the PCP. Default value: 0 (Range: 0 to 7)"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the 8 bit value mask. Default value: 0x07"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of VLAN information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the PCP. Default value: 0 (Range: 0 to 7)"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the 8 bit value mask. Default value: 0x07"""

    def get(self) -> Token[GetDataAttr]:
        """Get the VLAN PCP settings for the VLAN filter.

        :return: the VLAN PCP settings for the VLAN filter
        :rtype: PEF_VLANPCP.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._vlan_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set the VLAN PCP settings for the VLAN filter.

        :param use: specifies the use of VLAN PCP information
        :type use: OnOff
        :param value: specifying the value of the PCP. Default value: 0 (Range: 0 to 7)
        :type value: int
        :param mask: specifying the 8 bit value mask. Default value: 0x07
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._vlan_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """VLAN PCP is not used for the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """VLAN PCP is used for the filter."""


@register_command
@dataclass
class PEF_MPLSSETTINGS:
    """
    Basic mode only. Defines what filter action is performed on the MPLS header.
    """

    code: typing.ClassVar[int] = 1710
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of MPLS information"""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action on MPLS information"""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of MPLS information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action on MPLS information"""

    def get(self) -> Token[GetDataAttr]:
        """Get the filter action settings on the MPLS header.

        :return: the filter action settings on the MPLS header
        :rtype: PEF_MPLSSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set the filter action settings on the MPLS header.

        :param use: specifies the use of MPLS information
        :type use: FilterUse
        :param action: specifies specifies if MPLS information is expected
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_MPLSLABEL:
    """
    Basic mode only. Defines the MPLS label settings for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1711
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of MPLS information."""
        value: int = field(XmpInt())
        """integer, specifying the 20 bit value of the label. Default value: 0."""
        mask: Hex = field(XmpHex(size=3))
        """three hex bytes, specifying the 20 bit value of the label. Default value: 0x0FFFFF,"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of MPLS information."""
        value: int = field(XmpInt())
        """integer, specifying the 20 bit value of the label. Default value: 0."""
        mask: Hex = field(XmpHex(size=3))
        """three hex bytes, specifying the 20 bit value of the label. Default value: 0x0FFFFF,"""

    def get(self) -> Token[GetDataAttr]:
        """Get the MPLS label settings for the filter.

        :return: the MPLS label settings for the filter
        :rtype: PEF_MPLSLABEL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set the MPLS label settings for the filter.

        :param use: specifies the use of MPLS label information.
        :type use: OnOff
        :param value: specifying the 20-bit value of the label. Default value: 0.
        :type value: int
        :param mask: specifying the 20-bit value of the label. Default value: 0x0FFFFF,
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """The MPLS label is not used by the filter"""

    set_on = functools.partialmethod(set, OnOff.ON)
    """The MPLS label is used by the filter"""


@register_command
@dataclass
class PEF_MPLSTOC:
    """
    Basic mode only. Defines the MPLS TOC settings for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1712
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of MPLS TOC information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the MPLS TOC. Default value: 0 (Range: 0 to 7)."""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask for the value of the MPLS TOC. Default value: 0x07"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of MPLS TOC information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the MPLS TOC. Default value: 0 (Range: 0 to 7)."""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask for the value of the MPLS TOC. Default value: 0x07"""

    def get(self) -> Token[GetDataAttr]:
        """Get the MPLS TOC settings for the filter.

        :return: the MPLS TOC settings for the filter
        :rtype: PEF_MPLSTOC.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set the MPLS TOC settings for the filter.

        :param use: specifies the use of MPLS TOC information
        :type use: OnOff
        :param value: specifying the value of the MPLS TOC. Default value: 0 (Range: 0 to 7).
        :type value: int
        :param mask: specifying the filter mask for the value of the MPLS TOC. Default value: 0x07
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """The MPLS TOC is not used by the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """The MPLS TOC is used by the filter."""


@register_command
@dataclass
class PEF_L3USE:
    """
    Basic mode only. Defines what Layer 3 protocols that are present and may be used
    for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1713
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: L3Present = field(XmpByte())
        """coded byte, specifies the presence of Layer 3 protocols:"""

    class SetDataAttr(RequestBodyStruct):
        use: L3Present = field(XmpByte())
        """coded byte, specifies the presence of Layer 3 protocols:"""

    def get(self) -> Token[GetDataAttr]:
        """Get Layer 3 protocols settings for the filter.

        :return: Layer 3 protocols settings for the filter.
        :rtype: PEF_L3USE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: L3Present) -> Token[None]:
        """Set Layer 3 protocols settings for the filter.

        :param use: specifies the presence of Layer 3 protocols
        :type use: L3Present
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use))

    set_na = functools.partialmethod(set, L3Present.NA)
    """Set Layer 3 protocol presence to NA."""

    set_ip4 = functools.partialmethod(set, L3Present.IP4)
    """Set Layer 3 protocol presence to IPv4."""

    set_ip6 = functools.partialmethod(set, L3Present.IP6)
    """Set Layer 3 protocol presence to IPv6."""


@register_command
@dataclass
class PEF_IPV4SETTINGS:
    """
    Basic mode only. Defines what filter action is performed on the IPv4 header.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1714
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of IPv4 information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of IPv4 information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of IPv4 information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of IPv4 information."""

    def get(self) -> Token[GetDataAttr]:
        """Get the filter action settings on IPv4 header.

        :return: the filter action settings on IPv4 header
        :rtype: PEF_IPV4SETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set the filter action settings on IPv4 header.

        :param use: specifies the use of IPv4 information
        :type use: FilterUse
        :param action: specifies the action of IPv4 information
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_IPV4SRCADDR:
    """
    Basic mode only. Defines the IPv4 Source Address settings for the IPv4 filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1715
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 Source Address."""
        value: ipaddress.IPv4Address = field(XmpIPv4Address())
        """address, specifying the four bytes of the address. Default value: 0.0.0.0"""
        mask: Hex = field(XmpHex(size=4))
        """four hex bytes, specifying the filter mask of the value. Default value: 0xFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 Source Address."""
        value: ipaddress.IPv4Address = field(XmpIPv4Address())
        """address, specifying the four bytes of the address. Default value: 0.0.0.0"""
        mask: Hex = field(XmpHex(size=4))
        """four hex bytes, specifying the filter mask of the value. Default value: 0xFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the IPv4 Source Address settings for the IPv4 filter.

        :return: the IPv4 Source Address settings for the IPv4 filter
        :rtype: PEF_IPV4SRCADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: ipaddress.IPv4Address, mask: Hex) -> Token[None]:
        """Set the IPv4 Source Address settings for the IPv4 filter.

        :param use: specifies the use of IPv4 Source Address information
        :type use: OnOff
        :param value: specifying the four bytes of the address. Default value: 0.0.0.0
        :type value: ipaddress.IPv4Address
        :param mask: specifying the filter mask of the value. Default value: 0xFFFFFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv4 Source Address is not used by the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv4 Source Address is used by the filter."""


@register_command
@dataclass
class PEF_IPV4DESTADDR:
    """
    Basic mode only. Defines the IPv4 Destination Address settings for the IPv4 filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1716
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 Destination Address."""
        value: ipaddress.IPv4Address = field(XmpIPv4Address())
        """address, specifying the four bytes of the address. Default value: 0.0.0.0"""
        mask: Hex = field(XmpHex(size=4))
        """four hex bytes, specifying the filter mask of the value. Default value: 0xFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 Destination Address."""
        value: ipaddress.IPv4Address = field(XmpIPv4Address())
        """address, specifying the four bytes of the address. Default value: 0.0.0.0"""
        mask: Hex = field(XmpHex(size=4))
        """four hex bytes, specifying the filter mask of the value. Default value: 0xFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the IPv4 Destination Address settings for the IPv4 filter.

        :return: the IPv4 Destination Address settings for the IPv4 filter
        :rtype: PEF_IPV4DESTADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: ipaddress.IPv4Address, mask: Hex) -> Token[None]:
        """Set the IPv4 Destination Address settings for the IPv4 filter.

        :param use: specifies the use of IPv4 Destination Address information
        :type use: OnOff
        :param value: specifying the four bytes of the address. Default value: 0.0.0.0
        :type value: ipaddress.IPv4Address
        :param mask: specifying the filter mask of the value. Default value: 0xFFFFFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv4 Destination Address is not used by the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv4 Destination Address is used by the filter."""


@register_command
@dataclass
class PEF_IPV4DSCP:
    """
    Basic mode only. Defines if IPv4 DSCP/TOS settings used for the IPv4 filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1717
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the IPv4 DSCP/TOS in the upper 6 bits. value[7:2] = DSCP/TOS, value[1:0] = reserved (must be zero). Default value: 0"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask of the value in the upper 6 bits. mask[7:2] = DSCP/TOS mask, mask[1:0] = reserved (must be zero). Default value: 0xFC"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv4 information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the IPv4 DSCP/TOS in the upper 6 bits. value[7:2] = DSCP/TOS, value[1:0] = reserved (must be zero). Default value: 0"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask of the value in the upper 6 bits. mask[7:2] = DSCP/TOS mask, mask[1:0] = reserved (must be zero). Default value: 0xFC"""

    def get(self) -> Token[GetDataAttr]:
        """Get IPv4 DSCP/TOS settings for the filter.

        :return: IPv4 DSCP/TOS settings for the filter.
        :rtype: PEF_IPV4DSCP.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set IPv4 DSCP/TOS settings for the filter.

        :param use: specifies the use of IPv4 DSCP/TOS information.
        :type use: OnOff
        :param value: specifying the value of the IPv4 DSCP/TOS in the upper 6 bits. value[7:2] = DSCP/TOS, value[1:0] = reserved (must be zero). Default value: 0
        :type value: int
        :param mask: specifying the filter mask of the value in the upper 6 bits. mask[7:2] = DSCP/TOS mask, mask[1:0] = reserved (must be zero). Default value: 0xFC
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv4 DSCP/TOS is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv4 DSCP/TOS is used in the filter."""


@register_command
@dataclass
class PEF_IPV6SETTINGS:
    """
    Basic mode only. Defines what filter action is performed on the IPv6 header.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1718
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of IPv6 information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of IPv6 information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of IPv6 information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of IPv6 information."""

    def get(self) -> Token[GetDataAttr]:
        """Get filter action settings on the IPv6 header.

        :return: specifies the use of IPv6 header
        :rtype: PEF_IPV6SETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set filter action settings on the IPv6 header.

        :param use: specifies the use of IPv6 header
        :type use: FilterUse
        :param action: specifies the action of IPv6 header
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_IPV6SRCADDR:
    """
    Basic mode only. Defines the IPv6 Source Address settings for the IPv6 filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1719
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 Source Address."""
        value: ipaddress.IPv6Address = field(XmpIPv6Address())
        """16 hex bytes, specifying the address. Default : 0x00000000000000000000000000000000"""
        mask: Hex = field(XmpHex(size=16))
        """16 hex bytes, specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 Source Address."""
        value: ipaddress.IPv6Address = field(XmpIPv6Address())
        """16 hex bytes, specifying the address. Default : 0x00000000000000000000000000000000"""
        mask: Hex = field(XmpHex(size=16))
        """16 hex bytes, specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the IPv6 Source Address settings for the IPv6 filter.

        :return: the IPv6 Source Address settings for the IPv6 filter
        :rtype: PEF_IPV6SRCADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: ipaddress.IPv6Address, mask: Hex) -> Token[None]:
        """Set the IPv6 Source Address settings for the IPv6 filter.

        :param use: specifies the use of IPv6 Source Address information
        :type use: OnOff
        :param value: specifying the address. Default : 0x00000000000000000000000000000000
        :type value: ipaddress.IPv6Address
        :param mask: specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv6 Source Address is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv6 Source Address is used in the filter."""


@register_command
@dataclass
class PEF_IPV6DESTADDR:
    """
    Basic mode only. Defines the IPv6 Destination Address settings for the IPv6 filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1720
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 Destination Address."""
        value: ipaddress.IPv6Address = field(XmpIPv6Address())
        """16 hex bytes, specifying the address. Default : 0x00000000000000000000000000000000"""
        mask: Hex = field(XmpHex(size=16))
        """16 hex bytes, specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 Destination Address."""
        value: ipaddress.IPv6Address = field(XmpIPv6Address())
        """16 hex bytes, specifying the address. Default : 0x00000000000000000000000000000000"""
        mask: Hex = field(XmpHex(size=16))
        """16 hex bytes, specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the IPv6 Destination Address settings for the IPv6 filter.

        :return: IPv6 Destination Address settings for the IPv6 filter
        :rtype: PEF_IPV6DESTADDR.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: ipaddress.IPv6Address, mask: Hex) -> Token[None]:
        """Set the IPv6 Destination Address settings for the IPv6 filter.

        :param use:  specifies the use of IPv6 Destination Address information
        :type use: OnOff
        :param value: specifying the address. Default : 0x00000000000000000000000000000000
        :type value: ipaddress.IPv6Address
        :param mask: specifying the six first bytes of the address. Default value: 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv6 Destination Address is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv6 Destination Address is used in the filter."""


@register_command
@dataclass
class PEF_IPV6TC:
    """
    Basic mode only. Defines the IPv6 Traffic Class settings used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1721
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the IPv6 Traffic Class in the upper 6 bits. value[7:2] = IPv6 Traffic Class. value[1:0] = reserved (must be zero). Default value: 0"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask for the value in the upper 6 bits. mask[7:2] = IPv6 Traffic Class mask. mask[1:0] = reserved (must be zero). Default value: 0xFC"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of IPv6 information."""
        value: int = field(XmpByte())
        """byte, specifying the value of the IPv6 Traffic Class in the upper 6 bits. value[7:2] = IPv6 Traffic Class. value[1:0] = reserved (must be zero). Default value: 0"""
        mask: Hex = field(XmpHex(size=1))
        """hex byte, specifying the filter mask for the value in the upper 6 bits. mask[7:2] = IPv6 Traffic Class mask. mask[1:0] = reserved (must be zero). Default value: 0xFC"""

    def get(self) -> Token[GetDataAttr]:
        """Get IPv6 Traffic Class settings used for the filter.

        :return: IPv6 Traffic Class settings used for the filter
        :rtype: PEF_IPV6TC.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set IPv6 Traffic Class settings used for the filter.

        :param use: specifies the use of the IPv6 Traffic Class information.
        :type use: OnOff
        :param value: specifying the value of the IPv6 Traffic Class in the upper 6 bits. value[7:2] = IPv6 Traffic Class. value[1:0] = reserved (must be zero). Default value: 0
        :type value: int
        :param mask: specifying the filter mask for the value in the upper 6 bits. mask[7:2] = IPv6 Traffic Class mask. mask[1:0] = reserved (must be zero). Default value: 0xFC
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """IPv6 Traffic Class is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """IPv6 Traffic Class is used in the filter."""


@register_command
@dataclass
class PEF_UDPSETTINGS:
    """
    Basic mode only. Controls if UDP packet information is used for flow filtering.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1722
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of UDP information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies action use of UDP information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of UDP information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of UDP information."""

    def get(self) -> Token[GetDataAttr]:
        """Get filter settings on the UDP header.

        :return: filter action settings on the UDP header
        :rtype: PEF_UDPSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set filter settings on the UDP header.

        :param use: specifies the use of UDP information.
        :type use: FilterUse
        :param action: specifies the action of UDP information.
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_UDPSRCPORT:
    """
    Basic mode only. Defines UDP Source Port settings used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``

    """

    code: typing.ClassVar[int] = 1723
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of UDP Source Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the UDP Source Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of UDP Source Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the UDP Source Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get UDP Source Port settings used for the filter.

        :return: UDP Source Port settings used for the filter.
        :rtype: PEF_UDPSRCPORT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set UDP Source Port settings used for the filter.

        :param use: specifies the use of UDP Source Port information
        :type use: OnOff
        :param value: specifying the value of the UDP Source Port. Default value: 0
        :type value: int
        :param mask: specifying the filter mask for the value. Default value: 0xFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """UDP Source Port is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """UDP Source Port is used in the filter."""


@register_command
@dataclass
class PEF_UDPDESTPORT:
    """
    Basic mode only. Defines UDP Destination Port settings used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.
    """

    code: typing.ClassVar[int] = 1724
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of UDP Destination Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the UDP Destination Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of UDP Destination Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the UDP Destination Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get UDP Destination Port settings used for the filter.

        :return: UDP Destination Port settings used for the filter.
        :rtype: PEF_UDPDESTPORT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set UDP Destination Port settings used for the filter.

        :param use: specifies the use of UDP Destination Port information
        :type use: OnOff
        :param value: specifying the value of the UDP Destination Port. Default value: 0
        :type value: int
        :param mask: specifying the filter mask for the value. Default value: 0xFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """UDP Destination Port is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """UDP Destination Port is used in the filter."""


@register_command
@dataclass
class PEF_TCPSETTINGS:
    """
    Basic mode only. Defines if filtering on TCP information is used for flow
    filtering.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1725
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of TCP information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of TCP information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of TCP information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of TCP information."""

    def get(self) -> Token[GetDataAttr]:
        """Get filter action settings on the TCP header.

        :return: filter action settings on the TCP header.
        :rtype: PEF_TCPSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set filter action settings on the TCP header.

        :param use: specifies the use of TCP information.
        :type use: FilterUse
        :param action: specifies the action of TCP information.
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_TCPSRCPORT:
    """
    Basic mode only. Defines TCP Source Port settings used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1726
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TCP Source Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the TCP Source Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TCP Source Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the TCP Source Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get TCP Source Port settings used for the filter.

        :return: TCP Source Port settings used for the filter.
        :rtype: PEF_TCPSRCPORT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set TCP Source Port settings used for the filter.

        :param use:  specifies the use of TCP Source Port information
        :type use: OnOff
        :param value: specifies the value of the TCP Source Port. Default value: 0
        :type value: int
        :param mask: specifies the filter mask for the value. Default value: 0xFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """TCP Source Port is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """TCP Source Port is used in the filter."""


@register_command
@dataclass
class PEF_TCPDESTPORT:
    """
    Basic mode only. Defines TCP Destination Port settings used for the filter.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1727
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TCP Destination Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the TCP Destination Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TCP Destination Port information"""
        value: int = field(XmpInt())
        """integer , specifying the value of the TCP Destination Port. Default value: 0"""
        mask: Hex = field(XmpHex(size=2))
        """two hex bytes , specifying the filter mask for the value. Default value: 0xFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get TCP Destination Port settings used for the filter.

        :return: TCP Destination Port settings used for the filter.
        :rtype: PEF_TCPDESTPORT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: OnOff, value: int, mask: Hex) -> Token[None]:
        """Set TCP Destination Port settings used for the filter.

        :param use: specifies the use of TCP Destination Port information
        :type use: OnOff
        :param value: specifies the value of the TCP Destination Port. Default value: 0
        :type value: int
        :param mask: specifies the filter mask for the value. Default value: 0xFFFF
        :type mask: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, value=value, mask=mask))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """TCP Destination Port is not used in the filter."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """TCP Destination Port is used in the filter."""


@register_command
@dataclass
class PEF_ANYSETTINGS:
    """
    Basic mode only. Defines if filtering on ANY field in a packet is used for flow filtering.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1728
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of ANY field information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of ANY field information."""

    class SetDataAttr(RequestBodyStruct):
        use: FilterUse = field(XmpByte())
        """coded byte, specifies the use of ANY field information."""
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of ANY field information."""

    def get(self) -> Token[GetDataAttr]:
        """Get the settings of filtering state on ANY field in a packet.

        :return: settings of filtering state on ANY field in a packet.
        :rtype: PEF_ANYSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, use: FilterUse, action: InfoAction) -> Token[None]:
        """Set the settings of filtering state on ANY field in a packet.

        :param use: specifies the use of ANY field information.
        :type use: FilterUse
        :param action:  specifies the action of ANY field information.
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], use=use, action=action))


@register_command
@dataclass
class PEF_ANYCONFIG:
    """
    Basic mode only. Defines the ANY field filter configuration. The "ANY field"
    filter will match 6 consecutive bytes in the incoming packets at a programmable
    offset. Applying a mask, allows to only filter based on selected bits within the
    6 bytes.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1729
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        position: int = field(XmpByte())
        """byte, specifies the start position of the ANY field. Default value: 0, Range:0-127"""
        value: Hex = field(XmpHex(size=6))
        """6 hex bytes, specifying the six bytes of the field. Default value: 0x000000000000"""
        mask: Hex = field(XmpHex(size=6))
        """6 hex bytes, specifying the six bytes of the field. Default value: 0xFFFFFFFFFFFF"""

    class SetDataAttr(RequestBodyStruct):
        position: int = field(XmpByte())
        """byte, specifies the start position of the ANY field. Default value: 0, Range:0-127"""
        value: Hex = field(XmpHex(size=6))
        """6 hex bytes, specifying the six bytes of the field. Default value: 0x000000000000"""
        mask: Hex = field(XmpHex(size=6))
        """6 hex bytes, specifying the six bytes of the field. Default value: 0xFFFFFFFFFFFF"""

    def get(self) -> Token[GetDataAttr]:
        """Get the ANY field filter configuration.

        :return: the ANY field filter configuration
        :rtype: PEF_ANYCONFIG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, position: int, value: Hex, mask: Hex) -> Token[None]:
        """Set the ANY field filter configuration

        :param position: specifies the start position of the ANY field. Default value: 0, Range:0-127
        :type position: int
        :param value: specifying the six bytes of the field. Default value: 0x000000000000
        :type value: str
        :param mask: specifying the six bytes of the field. Default value: 0xFFFFFFFFFFFF
        :type mask: str
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], position=position, value=value, mask=mask))


@register_command
@dataclass
class PEF_TPLDSETTINGS:
    """
    Defines if filtering on TPLD field in a packet is used for flow filtering. The
    TPLD filter allows filtering based on the Xena TPLD ID. The TPLD
    ID is meta data, which can be inserted into the Ethernet packets by Xena traffic
    generators. For each flow filter, can the filter be based on 16 TPLD ID values.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1730
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of TPLD information."""

    class SetDataAttr(RequestBodyStruct):
        action: InfoAction = field(XmpByte())
        """coded byte, specifies the action of TPLD information."""

    def get(self) -> Token[GetDataAttr]:
        """Get the settings of filtering on TPLD field in a packet.

        :return: the settings of filtering on TPLD field in a packet.
        :rtype: PEF_TPLDSETTINGS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, action: InfoAction) -> Token[None]:
        """Set the settings of filtering on TPLD field in a packet.

        :param action: specifies the action of TPLD information.
        :type action: InfoAction
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], action=action))


@register_command
@dataclass
class PEF_TPLDCONFIG:
    """
    Defines the TPLD filter configuration. There are only 16 TPLD filter, thus the index values are from 0 to 15.

    .. note::

        For SET, the only allowed ``_filter_type`` is ``shadow-copy``.

    """

    code: typing.ClassVar[int] = 1731
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType
    _test_payload_filter_index: int

    class GetDataAttr(ResponseBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TPLD field information."""
        id: int = field(XmpInt())
        """int, specifies the TPLD ID. Range: 0-2015, Default value: 0"""

    class SetDataAttr(RequestBodyStruct):
        use: OnOff = field(XmpByte())
        """coded byte, specifies the use of TPLD field information."""
        id: int = field(XmpInt())
        """int, specifies the TPLD ID. Range: 0-2015, Default value: 0"""

    def get(self) -> Token[GetDataAttr]:
        """Get the TPLD filter configuration.

        :return: the use of TPLD field information, and the TPLD ID. Range: 0-2015, Default value: 0
        :rtype: PEF_TPLDCONFIG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._test_payload_filter_index]))

    def set(self, use: OnOff, id: int) -> Token[None]:
        """Set the TPLD filter configuration.

        :param use: specifies the use of TPLD field information
        :type use: OnOff
        :param id: specifies the TPLD ID. Range: 0-2015, Default value: 0
        :type id: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type, self._test_payload_filter_index], use=use, id=id))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """The TPLD information is not used."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """The TPLD information is used."""


@register_command
@dataclass
class PEF_ISSHADOWDIRTY:
    """
    Get shadow filter status (if shadow is in sync with working copy or not).
    """

    code: typing.ClassVar[int] = 1734
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        is_in_sync: YesNo = field(XmpByte())
        """coded byte, if shadow is in sync with working copy or not."""

    def get(self) -> Token[GetDataAttr]:
        """Get shadow filter status.

        :return: if shadow is in sync with working copy or not.
        :rtype: PEF_ISSHADOWDIRTY.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PEF_CANCEL:
    """Undo updates to shadow filter settings, sets dirty false."""

    code: typing.ClassVar[int] = 1735
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Undo updates to shadow filter settings, sets dirty false.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex]))


@register_command
@dataclass
class PEF_VALUE:
    """
    This command is valid only for ``Extended filter mode`` (check :class:`PEF_MODE`).

    Defines the byte values that can be matched if selected by :class:`PEF_MASK`.

    If ``<protocol_segment_index> = 0`` the maximum number of match value
    bytes that can be set is determined by the total length of the protocol segments
    specified with PEF_PROTOCOL.

    E.g. if PEF_PROTOCOL is set to ETHERNET then only
    12 bytes can be set. In order to set the full 128 bytes, either specify a
    detailed protocol segment list, or use the raw protocol segment type. This specifies 12 + 116 = 128 bytes.

    If ``<protocol_segment_index> != 0`` only the bytes covered by that segment are manipulated,
    so if PEF_PROTOCOL is set to ``ETHERNET VLAN ETHERTYPE eCPRI`` then ``<protocol_segment_index> = 4`` selects the 8
    bytes of the eCPRI header starting at byte position (12 + 2 + 4) = 18.

    For ``set`` command where fewer value bytes are provided than specified by the protocol segment, those unspecified bytes are set to zero.

    The ``get`` command always returns the number of bytes specified by the protocol segment.
    """

    code: typing.ClassVar[int] = 1777
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType
    _protocol_segment_index: int

    class GetDataAttr(ResponseBodyStruct):
        value: Hex = field(XmpHex())
        """list of hex bytes, the raw bytes comprising the packet header."""

    class SetDataAttr(RequestBodyStruct):
        value: Hex = field(XmpHex())
        """list of hex bytes, the raw bytes comprising the packet header."""

    def get(self) -> Token[GetDataAttr]:
        """Get the byte values that can be matched if selected by :class:`PEF_MASK`.

        :return: the byte values that can be matched if selected by :class:`PEF_MASK`
        :rtype: PEF_VALUE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, value: Hex) -> Token[None]:
        """Set the byte values that can be matched if selected by :class:`PEF_MASK`.

        :param value: the raw bytes comprising the packet header
        :type value: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], value=value))


@register_command
@dataclass
class PEF_MASK:
    """
    This command is valid only for ``Extended filter mode`` (check :class:`PEF_MODE`).

    Defines the mask byte values that select the values specified by :class:`PEF_VALUE`.

    For a chosen ``<protocol_segment_index>`` the first byte in the value masks the
    first byte of the corresponding :class:`PEF_VALUE` and so on.

    If ``<protocol_segment_index> = 0`` the maximum number of match value
    bytes that can be set is determined by the total length of the protocol segments
    specified with PEF_PROTOCOL`.

    E.g. if PEF_PROTOCOL is set to ETHERNET then only
    12 bytes can be set. In order to set the full 128 bytes, either specify a
    detailed protocol segment list, or use the raw protocol segment type. This specifies 12 + 116 = 128 bytes.

    If ``<protocol_segment_index> != 0`` only the bytes covered by that segment are manipulated,
    so if PEF_PROTOCOL is set to ``ETHERNET VLAN ETHERTYPE eCPRI`` then ``<protocol_segment_index> = 4`` selects the 8
    bytes of the eCPRI header starting at byte position (12 + 2 + 4) = 18.

    ``get/set`` semantics are similar to :class:`PEF_VALUE`.
    """

    code: typing.ClassVar[int] = 1778
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType
    _protocol_segment_index: int

    class GetDataAttr(ResponseBodyStruct):
        masks: Hex = field(XmpHex())

    class SetDataAttr(RequestBodyStruct):
        masks: Hex = field(XmpHex())

    def get(self) -> Token[GetDataAttr]:
        """Get the mask byte values that select the values specified by :class:`PEF_VALUE`.

        :return: the mask byte values that select the values specified by :class:`PEF_VALUE`.
        :rtype: PEF_MASK.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, masks: Hex) -> Token[None]:
        """Set the mask byte values that select the values specified by :class:`PEF_VALUE``.

        :param masks: mask byte values
        :type masks: Hex
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], masks=masks))


@register_command
@dataclass
class PEF_PROTOCOL:
    """
    This command is valid only for ``Extended filter mode`` (check :class:`PEF_MODE`).

    Defines the sequence of protocol segments that can be
    matched. The total length of the specified segments cannot exceed 128 bytes. If
    an existing sequence of segments is changed (using PEF_PROTOCOL) the underlying
    value and mask bytes remain unchanged, even though the semantics of those bytes
    may have changed. However, if the total length, in bytes, of the segments is
    reduced, then the excess bytes of value and mask are set to zero. I.e. to update
    an existing filter, you must first correct the list of segments (using
    PEF_PROTOCOL) and subsequently update the filtering value (using :class:`PEF_VALUE`) and filtering mask (:class:`PEF_MASK`).
    """

    code: typing.ClassVar[int] = 1779
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        segment_list: typing.List[ProtocolOption] = field(XmpSequence(types_chunk=[XmpByte()]))
        """list of bytes, specifying the list of protocol segment types in the order they are expected in a frame. First segment type must be ETHERNET;
        the following can be chosen freely.
        """

    class SetDataAttr(RequestBodyStruct):
        segment_list: typing.List[ProtocolOption] = field(XmpSequence(types_chunk=[XmpByte()]))
        """list of bytes, specifying the list of protocol segment types in the order they are expected in a frame. First segment type must be ETHERNET;
        the following can be chosen freely.
        """

    def get(self) -> Token[GetDataAttr]:
        """Get the sequence of protocol segments that can be matched.

        :return: the sequence of protocol segments that can be matched.
        :rtype: PEF_PROTOCOL.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, segment_list: typing.List[ProtocolOption]) -> Token[None]:
        """Set the sequence of protocol segments that can be matched.

        :param segment_list: specifying the list of protocol segment types in the order they are expected in a frame.
            First segment type must be ``ETHERNET``; the following can be chosen freely.
        :type segment_list: typing.List[ProtocolOption]
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], segment_list=segment_list))


@register_command
@dataclass
class PEF_MODE:
    """Control the filter mode."""

    code: typing.ClassVar[int] = 1780
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _flow_xindex: int
    _filter_type: FilterType

    class GetDataAttr(ResponseBodyStruct):
        mode: FilterMode = field(XmpByte())
        """integer, the mode of the filter."""

    class SetDataAttr(RequestBodyStruct):
        mode: FilterMode = field(XmpByte())
        """integer, the mode of the filter."""

    def get(self) -> Token[GetDataAttr]:
        """Get the filter mode.

        :return: the filter mode
        :rtype: PEF_MODE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type]))

    def set(self, mode: FilterMode) -> Token[None]:
        """Set the filter mode.

        :param mode: the mode of the filter.
        :type sid: FlowMode
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._flow_xindex, self._filter_type], mode=mode))

    set_basic = functools.partialmethod(set, FilterMode.BASIC)
    """Set the filter mode to Basic."""

    set_extended = functools.partialmethod(set, FilterMode.EXTENDED)
    """Set the filter mode to Extended."""




from __future__ import annotations
from abc import (
    abstractmethod,
    ABC
)
from io import BytesIO
import struct

from typing import (
    Any,
    Callable,
    Generic,
    Protocol,
    Type,
    TypeVar,
)

from typing_extensions import (
    NoReturn,
    Self,
)

from .field import FieldSpecs
from .exceptions import FirmwareVersionError

# region Types

GenericType = TypeVar("GenericType")


class SetInstance(Protocol):
    _buffer: BytesIO


class GetInstance(Protocol):
    _buffer: memoryview
    _stencil: tuple[tuple[str, int], ...]

# endregion


class FieldDescriptor(ABC, Generic[GenericType]):
    '''
    Descriptor representing getter and setter of the field
    '''
    __slots__ = ("idx", "specs", "public_name",)

    def __init__(self: Self, idx: int, specs: FieldSpecs) -> None:
        # Will be called from the Meta class
        self.idx = idx
        self.specs = specs

    def __set_name__(self, owner, name: str) -> None:
        self.public_name = name

    @abstractmethod
    def __set__(self: Self, instance, value: GenericType) -> None:
        # Executed at the runtimne
        raise NotImplementedError()

    @abstractmethod
    def __get__(self: Self, instance, cls) -> GenericType | Self:
        # Executed at the runtimne
        raise NotImplementedError()


class RequestFieldDescr(FieldDescriptor[GenericType]):
    __slots__ = ("to_xmp_context", "fmt")

    def __init__(self: Self, idx: int, specs: FieldSpecs, user_type: Type[Any]) -> None:
        super().__init__(idx, specs)
        self.fmt = self.specs.format()
        self.to_xmp_context: Callable[[Any], Any] = self.specs.get_context_formatter(user_type, False)

    def __set__(self: Self, instance: SetInstance, value: GenericType) -> None:
        """Transform values from Python to Bxmp and store them in to the buffer"""
        # Executed at the runtimne
        val_ = self.to_xmp_context(value)
        instance._buffer.write(self.specs.pack(self.fmt, val=val_))

    def __get__(self: Self, instance: SetInstance, cls) -> NoReturn | Self:
        # Executed at the runtimne
        if instance is None:
            return self
        raise RuntimeError from None


class ResponseFieldDescr(FieldDescriptor[GenericType]):
    __slots__ = ("to_py_context",)

    def __init__(self: Self, idx: int, specs: FieldSpecs, user_type: Type[Any]) -> None:
        super().__init__(idx, specs)
        self.to_py_context: Callable[[Any], Any] = self.specs.get_context_formatter(user_type, True)

    def __set__(self: Self, instance: GetInstance, value: GenericType) -> NoReturn:
        # Executed at the runtimne
        raise RuntimeError from None

    def __get__(self: Self, instance: GetInstance, cls) -> GenericType | Self:
        """Unpack values from the buffer and converting to expected type if required"""
        # Executed at the runtimne
        if instance is None:
            return self
        format_, offset_ = instance._stencil[self.idx]
        try:
            val_ = self.specs.unpack(
                format=format_,
                buffer=instance._buffer,
                offset=offset_,
            )
        except struct.error:
            raise FirmwareVersionError(
                cmd_name=instance.__class__.__qualname__,
                field_name=self.public_name,
                min_version=self.specs.min_version,
            ) from None
        else:
            return self.to_py_context(val_)



from __future__ import annotations
from dataclasses import dataclass
import typing
import functools
from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpInt,
    XmpSequence,
    XmpStr,
    Hex,
    XmpHex,
)
from .enums import (
    LinkTrainFrameLock,
    LinkTrainCmdResults,
    LinkTrainCmd,
    Layer1ConfigType,
    OnOff,
    FreyaLinkTrainingMode,
    FreyaAutonegMode,
    FreyaTecAbility,
    FreyaFECAbility,
    FreyaPauseAbility,
    AutoNegMode,
    AutoNegStatus,
    FreyaTechAbilityHCDStatus,
    FECMode,
    PauseMode,
    FreyaOutOfSyncPreset,
    TimeoutMode,
    LinkTrainingStatusMode,
    LinkTrainingStatus,
    LinkTrainingFailureType,
    Layer1Control,
    Layer1Opcode,
    FreyaPCSVariant,
    FreyaTecAbilityHCD,
)


@register_command
@dataclass
class PL1_AUTONEGINFO:
    """
    .. versionadded:: 2.0

    .. warning::

        Still in beta mode. Subjected to changes

    Get L1 auto-negotiation information. Information is split into a number of pages.

    """

    code: typing.ClassVar[int] = 385
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _page_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        rx_link_codeword_count: int = field(XmpInt(signed=False))
        """received number of Link Code Words (Base Pages)."""
        rx_next_page_message_count: int = field(XmpInt(signed=False))
        """received number of Next Pages - Message Pages."""
        rx_next_page_unformatted_count: int = field(XmpInt(signed=False))
        """received number of Nex Pages - Unformatted Pages."""
        tx_link_codeword_count: int = field(XmpInt(signed=False))
        """transmitted number of Link Code Words (Base Pages)."""
        tx_next_page_message_count: int = field(XmpInt(signed=False))
        """transmitted number of Next Pages - Message Pages."""
        tx_next_page_unformatted_count: int = field(XmpInt(signed=False))
        """transmitted number of Nex Pages - Unformatted Pages."""
        negotiation_hcd_fail_count: int = field(XmpInt(signed=False))
        """number of negotiation HCD (Highest Common Denominator) failures."""
        negotiation_fec_fail_count: int = field(XmpInt(signed=False))
        """number of negotiation FEC failures."""
        negotiation_loss_of_sync_count: int = field(XmpInt(signed=False))
        """number of negotiation Loss of Sync failures."""
        negotiation_timeout_count: int = field(XmpInt(signed=False))
        """number of negotiation timeouts."""
        negotiation_success_count: int = field(XmpInt(signed=False))
        """number of negotiation successes."""
        duration_us: int = field(XmpInt(signed=False))
        """duration of the auto-negotiation in microseconds, from autoneg is enabled on the port to the negotiation is finished."""

    def get(self) -> Token[GetDataAttr]:
        """Get L1 auto-negotiation information. Information is split into a number of pages.

        :return: L1 auto-negotiation information
        :rtype: PL1_AUTONEGINFO.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._page_xindex]))


@register_command
@dataclass
class PL1_LINKTRAININFO:
    """
    .. versionadded:: 2.0

    .. warning::

        Still in beta mode. Subjected to changes

    Get L1 link training information. Information is per Serdes and split into a number of pages.
    """

    code: typing.ClassVar[int] = 386
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int
    _page_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        duration_us: int = field(XmpInt(signed=False))
        """duration of the auto-negotiation process in microseconds, from autoneg is enabled on the port to the negotiation is finished."""
        lock_lost_count: int = field(XmpInt(signed=False))
        """number of lost locks on auto-neg."""
        pre1_current_level: int = field(XmpInt(signed=False))
        """c(-1) current level."""
        pre1_rx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-1) received number of increment requests."""
        pre1_rx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-1) received number of decrement requests."""
        pre1_rx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-1) received number of maximum limits of coefficient and equalization requests reached."""
        pre1_rx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-1) received number of maximum limits of equalization requests reached."""
        pre1_rx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-1) received number of coefficients not supported."""
        pre1_rx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-1) received number of coefficients at limit."""
        pre1_tx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of increment requests."""
        pre1_tx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of decrement requests."""
        pre1_tx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of maximum limits of coefficient and equalization requests reached."""
        pre1_tx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of maximum limits of equalization requests reached."""
        pre1_tx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of coefficients not supported."""
        pre1_tx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-1) transmitted number of coefficients at limit."""
        main_current_level: int = field(XmpInt(signed=False))
        """c(0) current level."""
        main_rx_increment_req_count: int = field(XmpInt(signed=False))
        """c(0) received number of increment requests."""
        main_rx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(0) received number of decrement requests."""
        main_rx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(0) received number of maximum limits of coefficient and equalization requests reached."""
        main_rx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(0) received number of maximum limits of equalization requests reached."""
        main_rx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(0) received number of coefficients not supported."""
        main_rx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(0) received number of coefficients at limit."""
        main_tx_increment_req_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of increment requests."""
        main_tx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of decrement requests."""
        main_tx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of maximum limits of coefficient and equalization requests reached."""
        main_tx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of maximum limits of equalization requests reached."""
        main_tx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of coefficients not supported."""
        main_tx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(0) transmitted number of coefficients at limit."""
        post1_current_level: int = field(XmpInt(signed=False))
        """c(1) current level."""
        post1_rx_increment_req_count: int = field(XmpInt(signed=False))
        """c(1) received number of increment requests."""
        post1_rx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(1) received number of decrement requests."""
        post1_rx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(1) received number of maximum limits of coefficient and equalization requests reached."""
        post1_rx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(1) received number of maximum limits of equalization requests reached."""
        post1_rx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(1) received number of coefficients not supported."""
        post1_rx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(1) received number of coefficients at limit."""
        post1_tx_increment_req_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of increment requests."""
        post1_tx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of decrement requests."""
        post1_tx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of maximum limits of coefficient and equalization requests reached."""
        post1_tx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of maximum limits of equalization requests reached."""
        post1_tx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of coefficients not supported."""
        post1_tx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(1) transmitted number of coefficients at limit."""
        pre2_current_level: int = field(XmpInt(signed=False))
        """c(-2) current level."""
        pre2_rx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-2) received number of increment requests."""
        pre2_rx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-2) received number of decrement requests."""
        pre2_rx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-2) received number of maximum limits of coefficient and equalization requests reached."""
        pre2_rx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-2) received number of maximum limits of equalization requests reached."""
        pre2_rx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-2) received number of coefficients not supported."""
        pre2_rx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-2) received number of coefficients at limit."""
        pre2_tx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of increment requests."""
        pre2_tx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of decrement requests."""
        pre2_tx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of maximum limits of coefficient and equalization requests reached."""
        pre2_tx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of maximum limits of equalization requests reached."""
        pre2_tx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of coefficients not supported."""
        pre2_tx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-2) transmitted number of coefficients at limit."""
        pre3_current_level: int = field(XmpInt(signed=False))
        """c(-3) current level."""
        pre3_rx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-3) received number of increment requests."""
        pre3_rx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-3) received number of decrement requests."""
        pre3_rx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-3) received number of maximum limits of coefficient and equalization requests reached."""
        pre3_rx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-3) received number of maximum limits of equalization requests reached."""
        pre3_rx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-3) received number of coefficients not supported."""
        pre3_rx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-3) received number of coefficients at limit."""
        pre3_tx_increment_req_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of increment requests."""
        pre3_tx_decrement_req_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of decrement requests."""
        pre3_tx_coeff_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of maximum limits of coefficient and equalization requests reached."""
        pre3_tx_eq_limit_reached_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of maximum limits of equalization requests reached."""
        pre3_tx_coeff_not_supported_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of coefficients not supported."""
        pre3_tx_coeff_at_limit_count: int = field(XmpInt(signed=False))
        """c(-3) transmitted number of coefficients at limit."""
        prbs_total_bits_high: int = field(XmpInt(signed=False))
        """PRBS total bits (most significant 32-bit)."""
        prbs_total_bits_low: int = field(XmpInt(signed=False))
        """PRBS total bits  (least significant 32-bit)."""
        prbs_total_error_bits_high: int = field(XmpInt(signed=False))
        """PRBS total error bits (most significant 32-bit, only bit 15-0 should be used)."""
        prbs_total_error_bits_low: int = field(XmpInt(signed=False))
        """PRBS total error bits (least significant 32-bit)."""
        frame_lock: LinkTrainFrameLock = field(XmpInt(signed=False))
        """frame lock status of the local end."""
        remote_frame_lock: LinkTrainFrameLock = field(XmpInt(signed=False))
        """frame lock status of the remote end."""
        num_frame_errors: int = field(XmpInt(signed=False))

        num_overruns: int = field(XmpInt(signed=False))

        last_ic_received: int = field(XmpInt(signed=False))

        last_ic_sent: int = field(XmpInt(signed=False))

    def get(self) -> Token[GetDataAttr]:
        """Get L1 link training information. Information is per Serdes and split into a number of pages.

        :return: L1 link training information
        :rtype: PL1_LINKTRAININFO.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._page_xindex]))


@register_command
@dataclass
class PL1_LOG:
    """
    .. versionadded:: 2.0

    .. warning::

        Still in beta mode. Subjected to changes

    Return a log line of either AN or LT for the given Serdes. The log string line contains the latest 100 lines.
    """

    code: typing.ClassVar[int] = 387
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        log_string: str = field(XmpStr())

    def get(self) -> Token[GetDataAttr]:
        """Return a log line of ANLT of all serdes on a port. (latest 100 lines)

        :return: a log line from AN/LT for the given Serdes.
        :rtype: PL1_LOG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))



@register_command
@dataclass
class PL1_CFG_TMP:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Configure some L1 parameters.
    """

    code: typing.ClassVar[int] = 388
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    _serdes_xindex: int
    _type: Layer1ConfigType

    class GetDataAttr(ResponseBodyStruct):
        """Data structure of the get response.
        """
        values: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))

    class SetDataAttr(RequestBodyStruct):
        """Data structure of the set action.
        """
        values: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))

    def get(self) -> "Token[GetDataAttr]":
        """Get various L1 parameters

        :return: various L1 parameters
        :rtype: PL1_CFG_TMP.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._type]))

    def set(self, values: typing.List[int]) -> "Token":
        """Get various L1 parameters

        :param values: L1 parameters
        :type values: typing.List[int]
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._type], values=values))
    
@register_command
@dataclass
class PL1_LINKTRAIN_CMD:
    """
    .. versionadded:: 1.1

    .. warning::

        Still in beta mode. Subjected to changes

    Link training RPC. Issue link training commands on a given serdes and poll for status
    """

    code: typing.ClassVar[int] = 389
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        cmd: int = field(XmpByte())

        arg: int = field(XmpByte())

        result: LinkTrainCmdResults = field(XmpByte())

        flags: int = field(XmpByte())

    class SetDataAttr(RequestBodyStruct):
        cmd: LinkTrainCmd = field(XmpByte())

        arg: int = field(XmpByte())

    def get(self) -> Token[GetDataAttr]:
        """Get status of current command

        :return: 4 bytes: command, arg, result, flags
        :rtype: PL1_LINKTRAIN_CMD.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))

    def set(self, cmd: LinkTrainCmd, arg: int) -> Token[None]:
        """Issue a link train command (cmd, arg)

        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex], cmd=cmd, arg=arg))

@register_command
@dataclass
class PL1_CTRL:
    """
    .. versionadded:: 2.5

    The Signal Integrity feature offers the equivalent of an Equivalent Time oscilloscope trace of the RX PAM4 signal (later, also PAM2). The trace is done with the A/D converter in the GTM receiver also doing the data sampling / CDR function, i.e. the trace is taken after the RX equalizer.

    The HW characteristics of the Versal GTM used in Freya are: Trace length = 2000 samples, sample resolution = 7 bits 2's complement, i.e. range = -64..63.

    Using the sampled eye scan feature through CLI involves two steps:

    Trigger the acquisition of a trace (PL1_CTRL)

    Retrieve the trace data (PL1_GET_DATA)

    This command is a generic control function related to Layer 1 / SERDES. For now, only used for signal integrity scan.
    """

    code: typing.ClassVar[int] = 424
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int
    _func_xindex: Layer1Control

    class SetDataAttr(RequestBodyStruct):
        opcode: Layer1Opcode = field(XmpInt())

    def set(self, opcode: Layer1Opcode) -> Token[None]:

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._func_xindex], opcode=opcode))
    
@register_command
@dataclass
class PL1_GET_DATA:
    """
    .. versionadded:: 2.5

    The Signal Integrity feature offers the equivalent of an Equivalent Time oscilloscope trace of the RX PAM4 signal (later, also PAM2). The trace is done with the A/D converter in the GTM receiver also doing the data sampling / CDR function, i.e. the trace is taken after the RX equalizer.

    The HW characteristics of the Versal GTM used in Freya are: Trace length = 2000 samples, sample resolution = 7 bits 2’s complement, i.e. range = -64..63.

    Using the sampled eye scan feature through CLI involves two steps:

    Trigger the acquisition of a trace (PL1_CTRL)

    Retrieve the trace data (PL1_GET_DATA)

    This command is a generic function to retrieve dynamic data related to Layer 1 / SERDES. For now, only used for signal integrity scan.

    For ``func==0``, sampled eye scan:

    * ``result==0``: No data available.

        "No data available" means that either a scan was never started, an acquisition was started and in progress, or the acquired data has become too old (e.g. older than 500 ms). The acquisition time for a trace is in the very low ms-range. If ``result==0``, ``sweep_no`` and ``age_us`` are dummy (=0), and no additional data are returned.

    * ``result==1``: Data returned. In that case, the rest of the parameters apply:

        ``sweep_no``: per-SERDES trace acquisition counter: 1,2,3… Each trace can be returned multiple times, to different users, within its lifetime. A new trace acquisition is triggered with the PL1_CTRL command.

        ``age_us``: The “age” of the trace data in microseconds, i.e. the time from data acquisition from hardware was completed until the time the command reply data is generated.

        ``value``: The rest of the reply is a set of 16 bit signed 2-complement sample values. With present hardware, the range of each sample is -64..63. In XMP scripting, each sample value is represented as two bytes, msb first.

        With present implementation, 2006 sample values (4012 bytes) are returned.

        The first 6 sample values are so-called “sampled levels”: <p1> <p2> < p3> <m1> <m2> <m3>
    """

    code: typing.ClassVar[int] = 425
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int
    _func_xindex: Layer1Control

    class GetDataAttr(ResponseBodyStruct):
        result: int = field(XmpInt())
        """Data availability."""

        sweep_no: int = field(XmpInt())
        """per-SERDES trace acquisition counter."""

        age_us: int = field(XmpInt())
        """the age of the trace data in microseconds, i.e. the time from data acquisition from hardware was completed until the time the command reply data is generated."""

        value: typing.List[int] = field(XmpSequence(types_chunk=[XmpByte()]))
        """a set of 16 bit signed 2-complement sample values. With present hardware, the range of each sample is -64..63. In CLI scripting, each sample value is represented as two bytes, msb first."""

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex, self._func_xindex]))
    
@register_command
@dataclass
class PL1_PHYTXEQ_LEVEL:
    """
    .. versionadded:: 2.5
    
    Control and monitor the equalizer settings of the on-board PHY in the transmission direction (towards the transceiver cage).

    .. note::

        PL1_PHYTXEQ, PL1_PHYTXEQ_LEVEL, and PL1_PHYTXEQ_COEFF facilitate the configuration and retrieval of TX tap values, each offering a unique perspective. Modifications made with any of these parameters will result in updates to the read results across all of them.

    """

    code: typing.ClassVar[int] = 430
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value in dB/10, ranges from 0 to 71. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value in dB/10, ranges from 0 to 71. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value in dB/10, ranges from 0 to 187. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value in mV, ranges from 507 to 998."""
        post: int = field(XmpInt())
        """integer, post tap value in dB/10, ranges from 0 to 187 Default = 0 (neutral)"""
        
    class SetDataAttr(RequestBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value in dB/10, ranges from 0 to 71. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value in dB/10, ranges from 0 to 71. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value in dB/10, ranges from 0 to 187. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value in mV, ranges from 507 to 998."""
        post: int = field(XmpInt())
        """integer, post tap value in dB/10, ranges from 0 to 187 Default = 0 (neutral)"""

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))

    def set(self, pre3:int, pre2: int, pre: int, main: int, post: int) -> Token[None]:

        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex], pre3=pre3, pre2=pre2, pre=pre, main=main, post=post))
    
@register_command
@dataclass
class PL1_PHYTXEQ_COEFF:
    """
    .. versionadded:: 2.5
    
    Control and monitor the equalizer settings of the on-board PHY in the transmission direction (towards the transceiver cage).

    .. note::

        PL1_PHYTXEQ, PL1_PHYTXEQ_LEVEL, and PL1_PHYTXEQ_COEFF facilitate the configuration and retrieval of TX tap values, each offering a unique perspective. Modifications made with any of these parameters will result in updates to the read results across all of them.

    """

    code: typing.ClassVar[int] = 431
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value, negative, scaled by 1E3. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value, positive, scaled by 1E3. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value, negative, scaled by 1E3. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value, positive, scaled by 1E3. Default = 1000"""
        post: int = field(XmpInt())
        """integer, post tap value, negative, scaled by 1E3. Default = 0 (neutral)"""
        
    class SetDataAttr(RequestBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value, negative, scaled by 1E3. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value, positive, scaled by 1E3. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value, negative, scaled by 1E3. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value, positive, scaled by 1E3. Default = 1000"""
        post: int = field(XmpInt())
        """integer, post tap value, negative, scaled by 1E3. Default = 0 (neutral)"""

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))

    def set(self, pre3:int, pre2: int, pre: int, main: int, post: int) -> Token[None]:

        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex], pre3=pre3, pre2=pre2, pre=pre, main=main, post=post))    

@register_command
@dataclass
class PL1_AUTONEG_STATUS:
    """
    .. versionadded:: 2.5

    Returns received technology abilities, FEC abilities, pause abilities, HCD technology ability, FEC mode result, and pause mode result.
    """

    code: typing.ClassVar[int] = 432
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        mode: AutoNegMode = field(XmpInt())
        autoneg_state: AutoNegStatus = field(XmpInt())
        received_tech_abilities: Hex = field(XmpHex(size=8))
        received_fec_abilities: Hex = field(XmpHex(size=1))
        received_pause_mode: Hex = field(XmpHex(size=1))
        tech_ability_hcd_status: FreyaTechAbilityHCDStatus = field(XmpInt())
        tech_ability_hcd_value: FreyaTecAbilityHCD = field(XmpInt())
        fec_mode_result: FECMode = field(XmpInt())
        pause_mode_result: PauseMode = field(XmpInt())

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

@register_command
@dataclass
class PL1_AUTONEG_ABILITIES:
    """
    .. versionadded:: 2.5

    Return the supported technology abilities, FEC abilities, and pause abilities of the port.
    """

    code: typing.ClassVar[int] = 433
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        tech_abilities_supported: Hex = field(XmpHex(size=8))
        """supported technology abilities by the port. This returns a value in Hex of the format 0xHHHHHHHH (64 bits). Each bit corresponds to technology ability as shown below. A bit of 1 means the corresponding technology ability is supported by the port."""

        fec_modes_supported: Hex = field(XmpHex(size=1))
        """supported FEC modes by the port. This returns a value in Hex of the format 0xH (8 bits). Each bit corresponds to FEC mode as shown below. A bit of 1 means the corresponding FEC mode is supported by the port."""

        pause_modes_supported: Hex = field(XmpHex(size=1))
        """pause abilities supported by the port. This returns a value in Hex of the format 0xH (8 bits). Each bit corresponds to pause mode as shown below. A bit of 1 means the corresponding FEC mode is supported by the port."""

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))
    
@register_command
@dataclass
class PL1_PCS_VARIANT:
    """
    .. versionadded:: 2.5

    PCS variant configuration.
    """

    code: typing.ClassVar[int] = 434
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        variant: FreyaPCSVariant = field(XmpByte())

    class SetDataAttr(RequestBodyStruct):
        variant: FreyaPCSVariant = field(XmpByte())

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, variant: FreyaPCSVariant) -> Token[None]:

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, variant=variant))


@register_command
@dataclass
class PL1_AUTONEG_CONFIG:
    """
    .. versionadded:: 2.5

    Auto-negotiation configuration for Freya
    """

    code: typing.ClassVar[int] = 440
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        advertised_tech_abilities: Hex = field(XmpHex(size=8))
        advertised_fec_abilities: Hex = field(XmpHex(size=1))
        advertised_pause_mode: Hex = field(XmpHex(size=1))

    class SetDataAttr(RequestBodyStruct):
        advertised_tech_abilities: Hex = field(XmpHex(size=8))
        advertised_fec_abilities: Hex = field(XmpHex(size=1))
        advertised_pause_mode: Hex = field(XmpHex(size=1))

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, advertised_tech_abilities: Hex, advertised_fec_abilities: Hex, advertised_pause_mode: Hex) -> Token[None]:

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, advertised_tech_abilities=advertised_tech_abilities, advertised_fec_abilities=advertised_fec_abilities, advertised_pause_mode=advertised_pause_mode))
    
@register_command
@dataclass
class PL1_ANLT:
    """
    .. versionadded:: 2.5

    ANLT action
    """

    code: typing.ClassVar[int] = 441
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        an_mode: FreyaAutonegMode = field(XmpByte())

        lt_mode: FreyaLinkTrainingMode = field(XmpByte())

    class SetDataAttr(RequestBodyStruct):
        an_mode: FreyaAutonegMode = field(XmpByte())

        lt_mode: FreyaLinkTrainingMode = field(XmpByte())

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, an_mode: FreyaAutonegMode, lt_mode: FreyaLinkTrainingMode) -> Token[None]:

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, an_mode=an_mode, lt_mode=lt_mode))
    
    enable_an_only = functools.partialmethod(set, FreyaAutonegMode.ENABLED, FreyaLinkTrainingMode.DISABLED)
    """Enable Autoneg only.
    """

    enable_lt_auto_only = functools.partialmethod(set, FreyaAutonegMode.DISABLED, FreyaLinkTrainingMode.ENABLED_AUTO)
    """Enable Link Training (auto) only.
    """

    enable_lt_interactive_only = functools.partialmethod(set, FreyaAutonegMode.DISABLED, FreyaLinkTrainingMode.ENABLED_INTERACTIVE)
    """Enable Link Training (interactive) only.
    """

    enable_an_lt_auto = functools.partialmethod(set, FreyaAutonegMode.ENABLED, FreyaLinkTrainingMode.ENABLED_AUTO)
    """Enable Autoneg + Link Training (auto).
    """

    enable_an_lt_interactive = functools.partialmethod(set, FreyaAutonegMode.ENABLED, FreyaLinkTrainingMode.ENABLED_INTERACTIVE)
    """Enable Autoneg + Link Training (interactive).
    """

    disable_anlt = functools.partialmethod(set, FreyaAutonegMode.DISABLED, FreyaLinkTrainingMode.DISABLED)
    """Disable ANLT.
    """

@register_command
@dataclass
class PL1_PHYTXEQ:
    """
    .. versionadded:: 2.5
    
    Control and monitor the equalizer settings of the on-board PHY in the transmission direction (towards the transceiver cage).

    .. note::

        PL1_PHYTXEQ, PL1_PHYTXEQ_LEVEL, and PL1_PHYTXEQ_COEFF facilitate the configuration and retrieval of TX tap values, each offering a unique perspective. Modifications made with any of these parameters will result in updates to the read results across all of them.

    """

    code: typing.ClassVar[int] = 442
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value."""
        post: int = field(XmpInt())
        """integer, post tap value. Default = 0 (neutral)"""
        
    class SetDataAttr(RequestBodyStruct):
        pre3: int = field(XmpInt())
        """integer, pre3 tap value. Default = 0 (neutral)"""
        pre2: int = field(XmpInt())
        """integer, pre2 tap value. Default = 0 (neutral)"""
        pre: int = field(XmpInt())
        """integer, pre tap value. Default = 0 (neutral)"""
        main: int = field(XmpInt())
        """integer, main tap value."""
        post: int = field(XmpInt())
        """integer, post tap value. Default = 0 (neutral)"""

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))

    def set(self, pre3:int, pre2: int, pre: int, main: int, post: int) -> Token[None]:

        return Token(
            self._connection,
            build_set_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex], pre3=pre3, pre2=pre2, pre=pre, main=main, post=post))

@register_command
@dataclass
class PL1_LINKTRAIN_CONFIG:
    """
    .. versionadded:: 2.5

    Per-port link training settings
    """

    code: typing.ClassVar[int] = 443
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        oos_preset: FreyaOutOfSyncPreset = field(XmpByte())
        timeout_mode: TimeoutMode = field(XmpByte())

    class SetDataAttr(RequestBodyStruct):
        oos_preset: FreyaOutOfSyncPreset = field(XmpByte())
        timeout_mode: TimeoutMode = field(XmpByte())

    def get(self) -> Token[GetDataAttr]:

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, oos_preset: FreyaOutOfSyncPreset, timeout_mode: TimeoutMode) -> Token[None]:

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, oos_preset=oos_preset, timeout_mode=timeout_mode))
    
@register_command
@dataclass
class PL1_LINKTRAIN_STATUS:
    """
    Per-lane link training status
    """

    code: typing.ClassVar[int] = 444
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _serdes_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        mode: LinkTrainingStatusMode = field(XmpByte())
        """coded byte, link training mode"""
        status: LinkTrainingStatus = field(XmpByte())
        """coded byte, lane status."""
        failure: LinkTrainingFailureType = field(XmpByte())
        """coded byte, failure type."""

    def get(self) -> Token[GetDataAttr]:
        """Get link training status of a lane of a port.

        :return: link training status of a lane of a port, including mode, lane status, and failure type.
        :rtype: PP_LINKTRAINSTATUS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._serdes_xindex]))


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P4G_TIME_HIST_CONF,
    P4G_PAYLOAD_HIST_CONF,
    P4G_TRANSACTION_HIST_CONF,
    P4G_APP_TRANSACTION_HIST,
    P4G_RECALC_TIME_HIST,
    P4G_RECALC_PAYLOAD_HIST,
    P4G_RECALC_TRANSACTION_HIST,
)


class GConfigHistogram:
    """L47 Histogram Config"""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.time = P4G_TIME_HIST_CONF(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TIME_HIST_CONF
        """
        self.payload = P4G_PAYLOAD_HIST_CONF(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_PAYLOAD_HIST_CONF
        """
        self.transaction = P4G_TRANSACTION_HIST_CONF(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_TRANSACTION_HIST_CONF
        """


class GRecalculatesHistogram:
    """L47 Histogram Recalculation."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.time = P4G_RECALC_TIME_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_RECALC_TIME_HIST
        """
        self.payload = P4G_RECALC_PAYLOAD_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_RECALC_PAYLOAD_HIST
        """
        self.transaction = P4G_RECALC_TRANSACTION_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_RECALC_TRANSACTION_HIST
        """


class GHistogram:
    """L47 Histogram."""
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.transaction = P4G_APP_TRANSACTION_HIST(conn, module_id, port_id, group_idx)
        """
        Representation of P4G_APP_TRANSACTION_HIST
        """
        self.recalculates = GRecalculatesHistogram(conn, module_id, port_id, group_idx)
        """L47 Histogram Recalculation."""
        self.config = GConfigHistogram(conn, module_id, port_id, group_idx)
        """L47 Histogram Config"""



from __future__ import annotations
from typing import Type
from ._constants import CommandStatus


class XmpStatusException(Exception):
    __slots__ = ("err_code", "msg", "cmd",)

    err_code: CommandStatus | None
    msg: str
    cmd: str


class XmpNoConnectionError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOCONNECTIONS
        self.msg = "Chassis has no available connection slots"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNoLoggedOnError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTLOGGEDON
        self.msg = "No command can be submitted before logon"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNotReservedError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTRESERVED
        self.msg = "Parameter cannot be set because containing resource not reserved"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNotWritableError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTWRITABLE
        self.msg = "Parameter cannot be set because it is not in a writable state"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNotReadableError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTREADABLE
        self.msg = "Parameter is write-only"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNotValidError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTVALID
        self.msg = "Operation not valid in current state"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadHeaderError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADHEADER
        self.msg = "Invalid magic word"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadCommandError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADCOMMAND
        self.msg = "Invalid command"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadParameterError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADPARAMETER
        self.msg = "Invalid parameter code"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadModuleError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADMODULE
        self.msg = "Invalid module index"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadPortError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADPORT
        self.msg = "Invalid port index"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadIndexError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADINDEX
        self.msg = "Invalid parameter index"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadSizeError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADSIZE
        self.msg = "Invalid size of data"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpBadValueError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.BADVALUE
        self.msg = "Invalid value of data"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpFailedError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.FAILED
        self.msg = "Failed to perform operation"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpMemoryFailureError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.MEMORYFAILURE
        self.msg = "Failed to allocate memory"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNoPeLicenseError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOLICPE
        self.msg = "No free PE license"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNoFreePortLicenseError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOLICPORT
        self.msg = "No free Port license"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpNotSupportedError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.NOTSUPPORTED
        self.msg = "Feature not supported"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpPendingError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.PENDING
        self.msg = "Status return will wait until command has been executed"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpModuleOperationNotSupportedByChassisError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.MODULE_OPERATION_NOT_SUPPORTED_BY_CHASSIS
        self.msg = "module is not supported by chassis - e.g. because multi-image requires x64 OS."
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpXlsFailedError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.XLSFAILED
        self.msg = "Could not establish connection to Xena License Server"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpXlsDeniedError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.XLSDENIED
        self.msg = "Request for resource rejected by Xena License Server"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpXlsInvalidError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = CommandStatus.XLSINVALID
        self.msg = "Trying to run Valkyrie VE with VulcanVE-300 resource"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


class XmpUnknownError(XmpStatusException):
    def __init__(self, cmd) -> None:
        self.err_code = None
        self.msg = "Unknown XOA status"
        self.cmd = cmd
        super().__init__(f"{self.cmd}: {self.msg}")


__EXCEPTIONS_MAP = {
    CommandStatus.NOCONNECTIONS: XmpNoConnectionError,
    CommandStatus.NOTLOGGEDON: XmpNoLoggedOnError,
    CommandStatus.NOTRESERVED: XmpNotReservedError,
    CommandStatus.NOTWRITABLE: XmpNotWritableError,
    CommandStatus.NOTREADABLE: XmpNotReadableError,
    CommandStatus.NOTVALID: XmpNotValidError,
    CommandStatus.BADHEADER: XmpBadHeaderError,
    CommandStatus.BADCOMMAND: XmpBadCommandError,
    CommandStatus.BADPARAMETER: XmpBadParameterError,
    CommandStatus.BADMODULE: XmpBadModuleError,
    CommandStatus.BADPORT: XmpBadPortError,
    CommandStatus.BADINDEX: XmpBadIndexError,
    CommandStatus.BADSIZE: XmpBadSizeError,
    CommandStatus.BADVALUE: XmpBadValueError,
    CommandStatus.FAILED: XmpFailedError,
    CommandStatus.MEMORYFAILURE: XmpMemoryFailureError,
    CommandStatus.NOLICPE: XmpNoPeLicenseError,
    CommandStatus.NOLICPORT: XmpNoFreePortLicenseError,
    CommandStatus.NOTSUPPORTED: XmpNotSupportedError,
    CommandStatus.PENDING: XmpPendingError,
    CommandStatus.MODULE_OPERATION_NOT_SUPPORTED_BY_CHASSIS: XmpModuleOperationNotSupportedByChassisError,
    CommandStatus.XLSFAILED: XmpXlsFailedError,
    CommandStatus.XLSDENIED: XmpXlsDeniedError,
    CommandStatus.XLSINVALID: XmpXlsInvalidError
}


def get_status_error(cmd_status: CommandStatus | None) -> Type[XmpStatusException]:
    global __EXCEPTIONS_MAP
    if cmd_status is None:
        return XmpUnknownError
    return __EXCEPTIONS_MAP.get(cmd_status, XmpUnknownError)



import functools
from typing import TYPE_CHECKING
from xoa_driver.internals.commands import (
    P_DYNAMIC,
    P_TXRUNTLENGTH,
    P_RXRUNTLENGTH,
    P_RXRUNTLEN_ERRS,
    P_TXPREAMBLE_REMOVE,
    P_RXPREAMBLE_INSERT,

)
from xoa_driver.internals.utils import attributes as utils
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .bases.port_l23_genuine import BasePortL23Genuine


class FamilyF(BasePortL23Genuine):
    ...


class POdin10G1S2P(FamilyF):
    """L23 port on Odin-10G-1S-2P module.
    """
    ...


class POdin10G1S2P_b(FamilyF):
    """L23 port on Odin-10G-1S-2P[b] module.
    """
    ...


class POdin10G1S2P_c(FamilyF):
    """L23 port on Odin-10G-1S-2P[c] module.
    """
    ...


class POdin10G1S6P(FamilyF):
    """L23 port on Odin-10G-1S-6P module.
    """
    ...


class Runt:
    """Runt settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.tx_length = P_TXRUNTLENGTH(conn, module_id, port_id)
        """L23 port's TX runt length.
        
        :type: P_TXRUNTLENGTH
        """

        self.rx_length = P_RXRUNTLENGTH(conn, module_id, port_id)
        """L23 port's RX runt length.
        
        :type: P_RXRUNTLENGTH
        """

        self.has_length_errors = P_RXRUNTLEN_ERRS(conn, module_id, port_id)
        """L23 port's RX runt length errors..
        
        :type: P_RXRUNTLEN_ERRS
        """


class Preamble:
    """Preamble settings."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.tx_remove = P_TXPREAMBLE_REMOVE(conn, module_id, port_id)
        """L23 port's removal of preamble from outgoing packets.
        
        :type: P_TXPREAMBLE_REMOVE
        """

        self.rx_insert = P_RXPREAMBLE_INSERT(conn, module_id, port_id)
        """L23 port's insertion of preamble into incoming packets.
        
        :type: P_RXPREAMBLE_INSERT
        """


class POdin10G1S6P_b(FamilyF):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.runt = Runt(conn, module_id, port_id)
        """Runt settings."""

        self.preamble = Preamble(conn, module_id, port_id)
        """Preamble settiNgs."""

class POdin10G6S6P_a(FamilyF):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.runt = Runt(conn, module_id, port_id)
        """Runt settings."""

        self.preamble = Preamble(conn, module_id, port_id)
        """Preamble settiNgs."""


class POdin10G1S2PT(FamilyF):
    """L23 port on Odin-10G-1S-2P-T module.
    """
    ...


class POdin10G1S2P_d(FamilyF):
    """L23 port on Odin-10G-1S-2P[d] module.
    """
    ...


class POdin10G1S12P(FamilyF):
    """L23 port on Odin-10G-1S-12P module.
    """
    ...


class POdin40G2S2P(FamilyF):
    """L23 port on Odin-40G-2S-2P module.
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.dynamic = P_DYNAMIC(conn, module_id, port_id)
        """L23 port's dynamic traffic change.
        
        :type: P_DYNAMIC
        """

    on_dynamic_change = functools.partialmethod(utils.on_event, P_DYNAMIC)
    """Register a callback to the event that the port's dynamic traffic setting changes."""


class POdin40G2S2PB(FamilyF):
    """L23 port on Odin-40G-2S-2P-B module.
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.dynamic = P_DYNAMIC(conn, module_id, port_id)
        """L23 port's dynamic traffic change.
        
        :type: P_DYNAMIC
        """

    on_dynamic_change = functools.partialmethod(utils.on_event, P_DYNAMIC)
    """Register a callback to the event that the port's dynamic traffic setting changes."""



from __future__ import annotations


class XoaException(Exception):
    ...


class TransporterException(XoaException):
    ...


class XoaConnectionTimeoutError(XoaException):
    def __init__(self, host: str, port: int, seconds_timeout: int) -> None:
        self.host = host
        self.port = port
        self.seconds_timeout = seconds_timeout
        self.msg = f"Can't establish the connect to {host}:{port}, in period of {seconds_timeout} sec."
        super().__init__(self.msg)


class XoaConnectionError(XoaException):
    def __init__(self, host: str, port: int) -> None:
        self.host = host
        self.port = port
        self.msg = f"Can't connect to {host}:{port}"
        super().__init__(self.msg)


class RepeatedRequestID(TransporterException):
    def __init__(self, request_identifier: int,) -> None:
        self.msg = f"""
        Got repeated request id {request_identifier}.
        This is a bug of xenaserver returning the same request identifier twice.
        """
        self.request_identifier = request_identifier
        super().__init__(self.msg)


class XoaLostFuture(XoaException):
    def __init__(self, req_id: int, cls_name: str) -> None:
        self.req_id = req_id
        self.cls_name = cls_name
        self.msg = f"Command was sent but the response handler was not registered {req_id} {cls_name}."
        super().__init__(self.msg)



from __future__ import annotations
from dataclasses import dataclass
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpHex,
    Hex,
)
from .enums import ResourceAllocationMode


@register_command
@dataclass
class M4E_MODE:
    """
    Select resource allocation mode.
    """

    code: typing.ClassVar[int] = 850
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int

    class GetDataAttr(ResponseBodyStruct):
        mode: ResourceAllocationMode = field(XmpByte())
        """coded byte, resource allocation mode."""

    class SetDataAttr(RequestBodyStruct):
        mode: ResourceAllocationMode = field(XmpByte())
        """coded byte, resource allocation mode."""

    def get(self) -> Token[GetDataAttr]:
        """Get the resource allocation mode.

        :return: resource allocation mode
        :rtype: M4E_MODE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mode: ResourceAllocationMode) -> Token[None]:
        """Set the resource allocation mode.

        :param mode: resource allocation mode
        :type mode: ResourceAllocationMode
        """

        return Token(self._connection, build_set_request(self, module=self._module, mode=mode))

    set_simple = functools.partialmethod(set, ResourceAllocationMode.SIMPLE)
    """Set resource allocation mode to Simple."""

    set_advanced = functools.partialmethod(set, ResourceAllocationMode.ADVANCED)
    """Set resource allocation mode to Advanced."""


@register_command
@dataclass
class M4E_RESERVE:
    """
    Advanced mode only: Reserve a number of PEs so they later can be assigned to specific ports.
    """

    code: typing.ClassVar[int] = 851
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int

    class GetDataAttr(ResponseBodyStruct):
        mask: Hex = field(XmpHex(size=8))
        """eight hex bytes, bitmask of PEs to reserve"""

    class SetDataAttr(RequestBodyStruct):
        mask: Hex = field(XmpHex(size=8))
        """eight hex bytes, bitmask of PEs to reserve"""

    def get(self) -> Token[GetDataAttr]:
        """Get the PEs reserved.

        :return: the number of PEs reserved.
        :rtype: M4E_RESERVE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module))

    def set(self, mask: Hex) -> Token[None]:
        """Set PEs reserved.

        :param mask: bitmask of PEs to reserve
        :type mask: str
        """

        return Token(self._connection, build_set_request(self, module=self._module, mask=mask))
